#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (c) 2017 Jeremie DECOCK (http://www.jdhp.org)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
======================================
PyAI - Test functions for optimization
======================================

This package contains some classical test functions for optimization
algorithms.

.. currentmodule:: ealib.functions

Test functions
==============

.. toctree::

    optimize.functions.sphere

"""

# The following lines are inspired by https://github.com/scipy/scipy/blob/master/scipy/optimize/__init__.py

# from .unconstrained import *
# from .noise import *

# __all__ = [s for s in dir() if not s.startswith('_')]

#print("ealib.functions.__init__.py:", __all__)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (c) 2017,2018,2019 Jeremie DECOCK (http://www.jdhp.org)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
This module contains some classical test functions for unconstrained continuous
single-objective optimization.
"""

__all__ = ['sphere', 'Sphere', 'sphere1d', 'sphere2d', 'sphere3d', 'sphere10d', 'sphere100d',     # TODO
           'rosen', 'Rosenbrock', 'rosen2d',
           'himmelblau', 'Himmelblau', 'himmelblau2d',
           'rastrigin', 'Rastrigin', 'rastrigin2d',
           'easom', 'Easom', 'easom2d',
           'crossintray', 'Crossintray', 'crossintray2d',
           'holder', 'Holder', 'holder2d']

import numpy as np
from typing import Callable, Dict, List, Optional, Tuple, Union

# GENERIC OBJECTIVE FUNCTION ##################################################

class _ObjectiveFunction:
    """Generic *objective function*.

    TODO
    """
    def __init__(self) -> None:
        self.name: str = f"{self.__class__.__name__}"

        self._objective_function: Optional[Callable[[np.ndarray], Union[float, np.ndarray]]] = None
        self._gradient_function: Optional[Callable[[np.ndarray], np.ndarray]] = None    # TODO: use a generic numeric derivative function by default
        self._hessian_function: Optional[Callable[[np.ndarray], np.ndarray]] = None     # TODO: use a generic numeric derivative function by default

        self.reset_eval_counters()
        self.reset_eval_logs()
        self.do_eval_logs: bool = False

        self.noise: Optional[Callable[[np.ndarray, Union[float, np.ndarray]], Union[float, np.ndarray]]] = None

        self.ndim: Optional[int] = None
        self.bounds: Optional[np.ndarray] = None

        self.continuous: Optional[bool] = None

        self.translation_vector: Optional[np.ndarray] = None

        self.function_name: Optional[str] = None
        self.function_formula: Optional[str] = None

        # self.arg_min: Optional[np.ndarray] = None


    @property
    def stochastic(self) -> bool:
        return self.noise is not None

    @property
    def unimodal(self) -> bool:
        raise NotImplementedError

    @property
    def arg_min(self) -> Union[np.ndarray, Tuple[np.ndarray, ...]]:
        raise NotImplementedError


    def reset_eval_counters(self) -> None:
        # TODO: make an external Log (or Counter) class
        self.num_eval: int = 0
        self.num_gradient_eval: int = 0
        self.num_hessian_eval: int = 0


    def reset_eval_logs(self) -> None:
        # TODO: make an external Log class
        self.eval_logs_dict: Dict[str, List[np.ndarray]] = {'x': [], 'fx': []}  # TODO


    def __call__(self, x: np.ndarray) -> Union[float, np.ndarray]:
        """Evaluate one or several points.

        This function is a wrapper that does several boring task aside the
        evaluation of `func`: check arguments, log results, ...

        Parameters
        ----------
        func : callable object
            The function used to evaluate `x`.
        y : ndarray
            The 1D or 2D numpy array containing the points to evaluate.
            If `x` is a 2D array, the coordinates of each points are
            distributed along *the first dimension*.
            For instance, to evaluate the three 2D points (0,0), (1,1) and
            (2,2), `x` have to be coded as the following:
            `x = np.array([[0, 1, 2], [0, 1, 2]])`
            so that the first point is given by `x[:,0]`, the second point by
            `x[:,1]`, ... (this makes functions definition much simpler).

        Returns
        -------
        float or ndarray
            The results of the evaluation: a scalar if only one point has been
            evaluated or a 1D numpy array if several points have been
            evaluated.
        """
        # Check self._objective_function ########
        assert self._objective_function is not None
        assert callable(self._objective_function)

        # Check x shape #########################
        if x.ndim > 0:
            if x.shape[0] != self.ndim:
                raise Exception('Wrong number of dimension: x has {} rows instead of {}.'.format(x.shape[0], self.ndim))

        # Update the evaluations counter ########
        # TODO: make an external Log (or Counter) class
        if (x.ndim == 0) or (x.ndim == 1):
            self.num_eval += 1
        elif x.ndim == 2:
            self.num_eval += x.shape[1]
        else:
            raise Exception('Wrong number of dimension: x is a {} dimensions numpy array ; 1 or 2 dimensions are expected.'.format(x.ndim))

        # Apply translation #####################
        if self.translation_vector is not None:
            x_translated = (x.T - self.translation_vector).T
        else:
            x_translated = x

        # Eval x ################################
        y = self._objective_function(x_translated)

        # Apply noise ###########################
        if self.noise is not None:
            y = self.noise(x, y)

        # Update the evals log ##################
        # TODO: make an external Log class
        if self.do_eval_logs:
            if y.ndim == 0:
                self.eval_logs_dict['x'].append(x)       # TODO
            elif y.ndim == 1:
                self.eval_logs_dict['x'].extend(x.T)     # TODO
            else:
                raise Exception("Wrong output dimension.")

            if y.ndim == 0:
                self.eval_logs_dict['fx'].append(y)    # TODO
            elif y.ndim == 1:
                self.eval_logs_dict['fx'].extend(y)    # TODO
            else:
                raise Exception("Wrong output dimension.")

        return y


    def gradient(self, x: np.ndarray) -> np.ndarray:
        """
        The derivative (i.e. gradient) of the objective function.

        Parameters
        ----------
        x : array_like
            One dimension Numpy array of the point at which the derivative is to be computed
            or a two dimension Numpy array of points at which the derivatives are to be computed.

        Returns
        -------
        float or array_like
            gradient of the objective function at `x`.
        """
        # Check self._gradient_function #########
        assert self._gradient_function is not None
        assert callable(self._gradient_function)

        # Check x shape #########################
        if x.shape[0] != self.ndim:
            raise Exception('Wrong number of dimension: x has {} rows instead of {}.'.format(x.shape[0], self.ndim))

        # Update the evaluations counter ########
        # TODO: make an external Log (or Counter) class
        if x.ndim == 1:
            self.num_gradient_eval += 1
        elif x.ndim == 2:
            self.num_gradient_eval += x.shape[1]
        else:
            raise Exception('Wrong number of dimension: x is a {} dimensions numpy array ; 1 or 2 dimensions are expected.'.format(x.ndim))

        # Apply translation #####################
        if self.translation_vector is not None:
            x_translated = (x.T - self.translation_vector).T
        else:
            x_translated = x

        # Eval x ################################
        grad = self._gradient_function(x_translated)

        return grad


    def hessian(self, x: np.ndarray) -> np.ndarray:
        """
        The Hessian matrix of the objective function.
        
        Parameters
        ----------
        x : array_like
            1-D array of points at which the Hessian matrix is to be computed.

        Returns
        -------
        rosen_hess : ndarray
            The Hessian matrix of the objective function at `x`.
        """
        # Check self._gradient_function #########
        assert self._hessian_function is not None
        assert callable(self._hessian_function)

        # Check x shape #########################
        if x.shape[0] != self.ndim:
            raise Exception('Wrong number of dimension: x has {} rows instead of {}.'.format(x.shape[0], self.ndim))

        # Update the evaluations counter ########
        # TODO: make an external Log (or Counter) class
        if x.ndim == 1:
            self.num_hessian_eval += 1
        elif x.ndim == 2:
            self.num_hessian_eval += x.shape[1]
        else:
            raise Exception('Wrong number of dimension: x is a {} dimensions numpy array ; 1 or 2 dimensions are expected.'.format(x.ndim))

        # Apply translation #####################
        if self.translation_vector is not None:
            x_translated = (x.T - self.translation_vector).T
        else:
            x_translated = x

        # Eval x ################################
        hess = self._hessian_function(x_translated)

        return hess


    def __str__(self) -> str:
        name = r""

        if self.stochastic is not None:
            name += "stochastic "

        if self.function_name is not None:
            name += self.function_name
        else:
            name += self.__class__.__name__

        if self.function_formula is not None:
            name += ": " + self.function_formula

        return name

# SPHERE FUNCTION #############################################################

def sphere(x):
    r"""The Sphere function.

    The Sphere function is a famous **convex** function used to test the performance of optimization algorithms.
    This function is very easy to optimize and can be used as a first test to check an optimization algorithm.

    .. math::

        f(\boldsymbol{x}) = \sum_{i=1}^{n} x_{i}^2

    Global minimum:

    .. math::

        f(\boldsymbol{0}) = 0

    Search domain:

    .. math::

        \boldsymbol{x} \in \mathbb{R}^n

    .. image:: sphere_3d.png

    .. image:: sphere.png

    Example 
    -------

    To evaluate the single 2D point :math:`x = \begin{pmatrix} 0 \\ 0 \end{pmatrix}`:

    >>> sphere( np.array([0, 0]) )
    0.0

    The result should be :math:`f(x) = 0`.

    Example
    -------

    To evaluate the single 3D point :math:`x = \begin{pmatrix} 1 \\ 1 \\ 1 \end{pmatrix}`:

    >>> sphere( np.array([1, 1, 1]) )
    3.0

    The result should be :math:`f(x) = 3.0`.

    Example
    -------

    To evaluate multiple 2D points :math:`x_1 = \begin{pmatrix} 0 \\ 0 \end{pmatrix}`,
    :math:`x_2 = \begin{pmatrix} 1 \\ 1 \end{pmatrix}` and
    :math:`x_3 = \begin{pmatrix} 2 \\ 2 \end{pmatrix}` at once:

    >>> sphere( np.array([[0, 1, 2], [0, 1, 2]]) )
    ... # doctest: +NORMALIZE_WHITESPACE
    array([0., 2., 8.])

    The result should be :math:`f(x_1) = 0`, :math:`f(x_2) = 1` and :math:`f(x_3) = 8`.

    Parameters
    ----------
    x : array_like
        One dimension Numpy array of the point at which the Sphere function is to be computed
        or a two dimension Numpy array of points at which the Sphere function is to be computed.

    Returns
    -------
    float or array_like
        The value(s) of the Sphere function for the given point(s) `x`.
    
    See Also
    --------
    sphere_gradient, sphere_hessian
    """
    # Remark: `sum(x**2.0)` is equivalent to `np.sum(x**2.0, axis=0)` but only the latter works if x is a scallar (e.g. x = np.float(3)).
    return np.sum(x**2.0, axis=0)


def sphere_gradient(x):
    """
    The derivative (i.e. gradient) of the Sphere function.

    Parameters
    ----------
    x : array_like
        One dimension Numpy array of the point at which the derivative is to be computed
        or a two dimension Numpy array of points at which the derivatives are to be computed.

    Returns
    -------
    float or array_like
         gradient of the Sphere function at `x`.
    
    See Also
    --------
    sphere, sphere_hessian
    """
    return 2.0 * x


def sphere_hessian(x):
    """
    The Hessian matrix of the Sphere function.
    
    Parameters
    ----------
    x : array_like
        1-D array of points at which the Hessian matrix is to be computed.

    Returns
    -------
    rosen_hess : ndarray
        The Hessian matrix of the Sphere function at `x`.

    See Also
    --------
    sphere, sphere_gradient
    """
    return 2.0 * np.ones(x.shape)


class Sphere(_ObjectiveFunction):
    """
    TODO
    """
    def __init__(self, ndim: int):
        super().__init__()

        self.name = f"Sphere {ndim}D"

        self._objective_function = sphere
        self._gradient_function = sphere_gradient
        self._hessian_function = sphere_hessian

        self.ndim = ndim

        self.bounds = np.ones((2, self.ndim))    # TODO: take this or the transpose of this ?
        self.bounds[0,:] = -10.
        self.bounds[1,:] =  10.

        self.continuous = True

    @property
    def unimodal(self) -> bool:
        return True

    @property
    def arg_min(self) -> np.ndarray:
        return np.zeros(self.ndim)


sphere1d = Sphere(ndim=1)

sphere2d = Sphere(ndim=2)

sphere3d = Sphere(ndim=3)

sphere10d = Sphere(ndim=10)

sphere100d = Sphere(ndim=100)

# ROSENBROCK FUNCTION #########################################################

def rosen(x):
    r"""The (extended) Rosenbrock function.

    The Rosenbrock function is a famous **non-convex** function used to test
    the performance of optimization algorithms. The classical two-dimensional
    version of this function is **unimodal** but its *extended* :math:`n`-dimensional
    version (with :math:`n \geq 4`) is **multimodal** [SHANG06]_.

    .. math::

        f(\boldsymbol{x}) = \sum_{i=1}^{n-1} \left[100 \left( x_{i+1} - x_{i}^{2} \right)^{2} + \left( x_{i} - 1 \right)^2 \right]

    Global minimum:

    .. math::

        \min =
        \begin{cases}
            n = 2 & \rightarrow \quad f(1,1) = 0, \\
            n = 3 & \rightarrow \quad f(1,1,1) = 0, \\
            n > 3 & \rightarrow \quad f(\underbrace{1,\dots,1}_{n{\text{ times}}}) = 0 \\
        \end{cases}

    Search domain:

    .. math::

        \boldsymbol{x} \in \mathbb{R}^n

    The Rosenbrock has exactly one (global) minimum :math:`(\underbrace{1, \dots,
    1}_{n{\text{ times}}})^\top` for :math:`n \leq 3` and an additional *local*
    minimum for :math:`n \geq 4` near :math:`(-1, 1, 1, \dots, 1)^\top`.

    See http://www.mitpressjournals.org/doi/abs/10.1162/evco.2006.14.1.119
    (freely available at http://dl.acm.org/citation.cfm?id=1118014) and
    https://en.wikipedia.org/wiki/Rosenbrock_function#Multidimensional_generalisations
    for more information.

    See https://en.wikipedia.org/wiki/Rosenbrock_function and
    http://mathworld.wolfram.com/RosenbrockFunction.html for more information.

    The Rosenbrock function, its derivative (i.e. gradient) and its hessian matrix are also implemented in Scipy
    ([scipy.optimize.rosen](https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.rosen.html#scipy.optimize.rosen),
    [scipy.optimize.rosen_der](https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.rosen_der.html#scipy.optimize.rosen_der),
    [scipy.optimize.rosen_hess](https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.rosen_hess.html#scipy.optimize.rosen_hess) and
    [scipy.optimize.rosen_hess_prod](https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.rosen_hess_prod.html#scipy.optimize.rosen_hess_prod)).
    See [Scipy documentation](https://docs.scipy.org/doc/scipy/reference/optimize.html#rosenbrock-function) for more information.

    .. image:: rosenbrock_3d.png

    .. image:: rosenbrock.png

    Parameters
    ----------
    x : array_like
        One dimension Numpy array of the point at which the Rosenbrock function is to be computed
        or a two dimension Numpy array of points at which the Rosenbrock function is to be computed.

    Returns
    -------
    float or array_like
        The value(s) of the Rosenbrock function for the given point(s) `x`.

    Example
    -------

    To evaluate a single 2D point :math:`x = \begin{pmatrix} 0 \\ 0 \end{pmatrix}`:

    >>> rosen( np.array([0, 0]) )
    1.0

    The result should be :math:`f(x) = 1`.

    Example
    -------

    To evaluate a single 3D point :math:`x = \begin{pmatrix} 1 \\ 1 \\ 1 \end{pmatrix}`:

    >>> rosen( np.array([1, 1, 1]) )
    0.0

    The result should be :math:`f(x) = 0`.

    Example
    -------

    To evaluate multiple 2D points :math:`x_1 = \begin{pmatrix} 0 \\ 0 \end{pmatrix}`,
    :math:`x_2 = \begin{pmatrix} 1 \\ 1 \end{pmatrix}` and
    :math:`x_3 = \begin{pmatrix} 2 \\ 2 \end{pmatrix}` at once:

    >>> rosen( np.array([[0, 1, 2], [0, 1, 2]]) )
    ... # doctest: +NORMALIZE_WHITESPACE
    array([ 1., 0., 401.])

    The result should be :math:`f(x_1) = 1`, :math:`f(x_2) = 0` and :math:`f(x_3) = 401`.

    References
    ----------
    .. [SHANG06] `Shang, Y. W., & Qiu, Y. H. (2006). A note on the extended Rosenbrock function. Evolutionary Computation, 14(1), 119-126. <http://www.mitpressjournals.org/doi/abs/10.1162/evco.2006.14.1.119>`_
    """
    return np.sum(100.0*(x[1:] - x[:-1]**2.0)**2.0 + (1 - x[:-1])**2.0, axis=0)


class Rosenbrock(_ObjectiveFunction):
    """
    TODO
    """
    def __init__(self, ndim):
        super().__init__()

        self.name = f"Rosenbrock {ndim}D"

        self._objective_function = rosen

        self.ndim = ndim
        if self.ndim < 2: # TODO
            raise ValueError("The rosenbrock function is defined for solution spaces having at least 2 dimensions.")

        self.bounds = np.ones((2, self.ndim))    # TODO: take this or the transpose of this ?
        self.bounds[0,:] = -10. # TODO
        self.bounds[1,:] =  10. # TODO

        self.continuous = True

    @property
    def unimodal(self):
        return True if self.ndim < 4 else False

    @property
    def arg_min(self) -> np.ndarray:
        # https://en.wikipedia.org/wiki/Test_functions_for_optimization
        return np.ones(self.ndim)


rosen2d = Rosenbrock(ndim=2)


# HIMMELBLAU'S FUNCTION #######################################################

def himmelblau(x):
    r"""The Himmelblau's function.

    The Himmelblau's function is a two-dimensional **multimodal** function.

    .. math::

        f(x_1, x_2) = (x_1^2 + x_2 - 11)^2 + (x_1 + x_2^2 - 7)^2

    The function has four global minima:

    .. math::

        \begin{eqnarray}
            f(3, 2) = 0 \\
            f(-2.805118, 3.131312) = 0 \\
            f(-3.779310, -3.283186) = 0 \\
            f(3.584428, -1.848126) = 0
        \end{eqnarray}

    Search domain:

    .. math::

        \boldsymbol{x} \in \mathbb{R}^2

    It also has one local maximum at :math:`f(-0.270845, -0.923039) = 181.617`.

    The locations of all the minima can be found analytically (roots of cubic
    polynomials) but expressions are somewhat complicated.

    The function is named after David Mautner Himmelblau, who introduced it in
    *Applied Nonlinear Programming* (1972), McGraw-Hill, ISBN 0-07-028921-2.

    See https://en.wikipedia.org/wiki/Himmelblau%27s_function for more information.

    .. image:: himmelblau_3d.png

    .. image:: himmelblau.png

    Example
    -------

    To evaluate a single point :math:`x = \begin{pmatrix} 3 \\ 2 \end{pmatrix}`:

    >>> himmelblau( np.array([3, 2]) )
    0.0

    The result should be :math:`f(x) = 1`.

    Example
    -------

    To evaluate multiple points :math:`x_1 = \begin{pmatrix} 0 \\ 0 \end{pmatrix}`,
    :math:`x_2 = \begin{pmatrix} 1 \\ 1 \end{pmatrix}` and
    :math:`x_3 = \begin{pmatrix} 2 \\ 2 \end{pmatrix}` at once:

    >>> himmelblau( np.array([[0, 1, 2], [0, 1, 2]]) )
    ... # doctest: +NORMALIZE_WHITESPACE
    array([170.,  106.,   26.])

    The result should be :math:`f(x_1) = 170`, :math:`f(x_2) = 106` and :math:`f(x_3) = 26`.

    Parameters
    ----------
    x : array_like
        One dimension Numpy array of the point at which the Himmelblau's function is to be computed
        or a two dimension Numpy array of points at which the Himmelblau's function is to be computed.

    Returns
    -------
    float or array_like
        The value(s) of the Himmelblau's function for the given point(s) `x`.
    """
    assert x.shape[0] == 2, x.shape
    return (x[0]**2.0 + x[1] - 11.0)**2.0 + (x[0] + x[1]**2.0 - 7.0)**2.0


class Himmelblau(_ObjectiveFunction):
    """
    TODO
    """
    def __init__(self, ndim):
        super().__init__()

        self.name = f"Himmelblau {ndim}D"

        self._objective_function = himmelblau

        self.ndim = ndim
        if self.ndim != 2:
            raise ValueError("The himmelblau function is defined for solution spaces having 2 dimensions.")

        self.bounds = np.ones((2, self.ndim))    # TODO: take this or the transpose of this ?
        self.bounds[0,:] = -10. # TODO
        self.bounds[1,:] =  10. # TODO

        self.continuous = True

    @property
    def unimodal(self):
        return False

    @property
    def arg_min(self) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        # https://en.wikipedia.org/wiki/Test_functions_for_optimization
        return np.array([3.0, 2.0]), np.array([-2.805118, 3.131312]), np.array([-3.779310, -3.283186]), np.array([3.584428, -1.848126])


himmelblau2d = Himmelblau(ndim=2)


# RASTRIGIN FUNCTION ##########################################################

def rastrigin(x):
    r"""The Rastrigin function.

    The Rastrigin function is a famous **multimodal** function.
    Finding the minimum of this function is a fairly difficult problem due to
    its large search space and its large number of local minima.

    The classical two-dimensional version of this function has been introduced
    by L. A. Rastrigin in *Systems of extremal control* Mir, Moscow (1974).

    Its *generalized* :math:`n`-dimensional version has been proposed by H.
    Mühlenbein, D. Schomisch and J. Born in *The Parallel Genetic Algorithm as
    Function Optimizer* Parallel Computing, 17, pages 619–632, 1991.

    On an n-dimensional domain it is defined by:

    .. math::

        f(\boldsymbol{x}) = An + \sum_{i=1}^{n} \left[ x_{i}^{2} - A \cos(2 \pi x_{i}) \right]

    where :math:`A = 10`.

    Global minimum:

    .. math::

        f(\boldsymbol{0}) = 0

    Search domain:

    .. math::

        \boldsymbol{x} \in \mathbb{R}^n

    See https://en.wikipedia.org/wiki/Rastrigin_function for more information.

    .. image:: rastrigin_3d.png

    .. image:: rastrigin.png

    Example
    -------

    To evaluate a single 2D point :math:`x = \begin{pmatrix} 0 \\ 0 \end{pmatrix}`:

    >>> rastrigin( np.array([0, 0]) )
    0.0

    The result should be :math:`f(x) = 0`.

    Example
    -------

    To evaluate a single 3D point :math:`x = \begin{pmatrix} 1 \\ 1 \\ 1 \end{pmatrix}`:

    >>> rastrigin( np.array([0, 0, 0]) )
    0.0

    The result should be :math:`f(x) = 0`.

    Example
    -------

    To evaluate multiple 2D points :math:`x_1 = \begin{pmatrix} 0 \\ 0 \end{pmatrix}`,
    :math:`x_2 = \begin{pmatrix} 1 \\ 1 \end{pmatrix}` and
    :math:`x_3 = \begin{pmatrix} 2 \\ 2 \end{pmatrix}` at once:

    >>> rastrigin( np.array([[0, 1, 2], [0, 1, 2]]) )
    ... # doctest: +NORMALIZE_WHITESPACE
    array([ 1., 0., 401.])

    The result should be :math:`f(x_1) = 1`, :math:`f(x_2) = 0` and :math:`f(x_3) = 401`.

    Parameters
    ----------
    x : array_like
        One dimension Numpy array of the point at which the Rastrigin function is to be computed
        or a two dimension Numpy array of points at which the Rastrigin function is to be computed.

    Returns
    -------
    float or array_like
        The value(s) of the Rastrigin function for the given point(s) `x`.
    """
    A = 10.
    n = x.shape[0]
    return A * n + np.sum(x**2.0 - A * np.cos(2.0 * np.pi * x), axis=0)


class Rastrigin(_ObjectiveFunction):
    """
    TODO
    """
    def __init__(self, ndim):
        super().__init__()

        self.name = f"Rastrigin {ndim}D"

        self._objective_function = rastrigin

        self.ndim = ndim
        if self.ndim < 2: # TODO
            raise ValueError("The rastrigin function is defined for solution spaces having at least 2 dimensions.")

        self.bounds = np.ones((2, self.ndim))    # TODO: take this or the transpose of this ?
        self.bounds[0,:] = -10. # TODO
        self.bounds[1,:] =  10. # TODO

        self.continuous = True

    @property
    def unimodal(self):
        return False

    @property
    def arg_min(self) -> np.ndarray:
        # https://en.wikipedia.org/wiki/Test_functions_for_optimization
        return np.zeros(self.ndim)


rastrigin2d = Rastrigin(ndim=2)


# EASOM FUNCTION ##############################################################

def easom(x):
    r"""The Easom function.

    The Easom function is a 2 dimensions **unimodal** function.

    .. math::

        f(x_1, x_2) = -\cos(x_1) \cos(x_2) \exp \left( -\left[ (x_1-\pi)^2 + (x_2-\pi)^2 \right] \right)

    Global minimum:

    .. math::

        f(\pi, \pi) = -1

    Search domain:

    .. math::

        \boldsymbol{x} \in \mathbb{R}^2

    See https://www.sfu.ca/~ssurjano/easom.html for more information.

    .. image:: easom_3d.png

    .. image:: easom.png

    Example
    -------

    To evaluate a single 2D point :math:`x = \begin{pmatrix} 0 \\ 0 \end{pmatrix}`:

    >>> easom( np.array([np.pi, np.pi]) )
    -1.0

    The result should be :math:`f(x) = -1`.

    Example
    -------

    To evaluate multiple 2D points :math:`x_1 = \begin{pmatrix} \pi \\ \pi \end{pmatrix}`,
    :math:`x_2 = \begin{pmatrix} 0 \\ 0 \end{pmatrix}` and
    :math:`x_3 = \begin{pmatrix} 1 \\ 1 \end{pmatrix}` at once:

    >>> easom( np.array([[np.pi, 0, 1], [np.pi, 0, 1]]) )
    ... # doctest: +NORMALIZE_WHITESPACE
    array([-1., -2.67528799e-09, -3.03082341e-05])

    The result should be :math:`f(x_1) = -1`, :math:`f(x_2) \approx 0` and :math:`f(x_3) \approx 0`.

    Parameters
    ----------
    x : array_like
        One dimension Numpy array of the point at which the Easom function is to be computed
        or a two dimension Numpy array of points at which the Easom function is to be computed.

    Returns
    -------
    float or array_like
        The value(s) of the Easom function for the given point(s) `x`.
    """
    assert x.shape[0] == 2, x.shape
    return -np.cos(x[0]) * np.cos(x[1]) * np.exp(-((x[0]-np.pi)**2.0 + (x[1]-np.pi)**2.0))


class Easom(_ObjectiveFunction):
    """
    TODO
    """
    def __init__(self, ndim):
        super().__init__()

        self._objective_function = easom

        self.ndim = ndim
        if self.ndim != 2:
            raise ValueError("The easom function is defined for solution spaces having 2 dimensions.")

        self.bounds = np.ones((2, self.ndim))    # TODO: take this or the transpose of this ?
        self.bounds[0,:] = -10. # TODO
        self.bounds[1,:] =  10. # TODO

        self.continuous = True

    @property
    def unimodal(self):
        return True

    @property
    def arg_min(self) -> np.ndarray:
        # https://en.wikipedia.org/wiki/Test_functions_for_optimization
        return np.array([np.pi, np.pi])


easom2d = Easom(ndim=2)


# CROSS-IN-TRAY FUNCTION ######################################################

def crossintray(x):
    r"""The Cross-in-tray function.

    The Cross-in-tray function is a 2 dimensions **multimodal** function, with
    four global minima.

    .. math::

        f(x_1, x_2) = -0.0001 \left( \left| \sin(x_1) \sin(x_2) \exp \left( \left| 100 - \frac{\sqrt{x_1^2 + x_2^2}}{\pi} \right| \right)\right| + 1 \right)^{0.1}

    Global minima:

    .. math::

        \text{Min} =
        \begin{cases}
            f(1.34941, -1.34941)  &= -2.06261 \\
            f(1.34941, 1.34941)   &= -2.06261 \\
            f(-1.34941, 1.34941)  &= -2.06261 \\
            f(-1.34941, -1.34941) &= -2.06261 \\
        \end{cases}

    Search domain:

    .. math::

        -10 \leq x_1, x_2 \leq 10

    **References**: *Test functions for optimization* (Wikipedia):
    https://en.wikipedia.org/wiki/Test_functions_for_optimization.

    .. image:: cross_in_tray_3d.png

    .. image:: cross_in_tray.png

    Example
    -------

    To evaluate a single 2D point :math:`x = \begin{pmatrix} 0 \\ 0 \end{pmatrix}`:

    >>> crossintray( np.array([0, 0]) )
    -0.0001

    The result should be :math:`f(x) = -0.0001`.

    Example
    -------

    To evaluate multiple 2D points :math:`x_1 = \begin{pmatrix} 0 \\ 0 \end{pmatrix}`,
    :math:`x_2 = \begin{pmatrix} 1.34941 \\ 1.34941 \end{pmatrix}` and
    :math:`x_3 = \begin{pmatrix} -1.34941 \\ -1.34941 \end{pmatrix}` at once:

    >>> crossintray( np.array([[0, 1.34941, -1.34941], [0, 1.34941, -1.34941]]) )
    ... # doctest: +NORMALIZE_WHITESPACE
    array([ -0.0001,    -2.06261,  -2.06261])

    The result should be :math:`f(x_1) = -0.0001`, :math:`f(x_2) = -2.06261` and :math:`f(x_3) = -2.06261`.

    Parameters
    ----------
    x : array_like
        One dimension Numpy array of the point at which the Cross-in-tray function is to be computed
        or a two dimension Numpy array of points at which the Cross-in-tray function is to be computed.

    Returns
    -------
    float or array_like
        The value(s) of the Cross-in-tray function for the given point(s) `x`.
    """
    assert x.shape[0] == 2, x.shape
    return -0.0001 * (np.abs(np.sin(x[0]) * np.sin(x[1]) * np.exp( np.abs( 100.0 - np.sqrt(x[0]**2.0 + x[1]**2.0)/np.pi ))) + 1.0)**0.1


class Crossintray(_ObjectiveFunction):
    """
    TODO
    """
    def __init__(self, ndim):
        super().__init__()

        self._objective_function = crossintray

        self.ndim = ndim
        if self.ndim != 2:
            raise ValueError("The crossintray function is defined for solution spaces having 2 dimensions.")

        self.bounds = np.ones((2, self.ndim))    # TODO: take this or the transpose of this ?
        self.bounds[0,:] = -10.
        self.bounds[1,:] =  10.

        self.continuous = True

    @property
    def unimodal(self):
        return False

    @property
    def arg_min(self) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        # https://en.wikipedia.org/wiki/Test_functions_for_optimization
        return (np.array([1.34941, -1.34941]), np.array([1.34941, 1.34941]), np.array([-1.34941, 1.34941]), np.array([-1.34941, -1.34941]))


crossintray2d = Crossintray(ndim=2)


# HÖLDER TABLE FUNCTION #######################################################

def holder(x):
    r"""The Hölder table function.

    The Hölder table function is a 2 dimensions **multimodal** function, with
    four global minima.

    .. math::

        f(x_1, x_2) =
        -\left| \sin(x_1) \cos(x_2) \exp \left( \left| 1 - \frac{\sqrt{x_1^2 + x_2^2}}{\pi} \right| \right) \right|

    Global minima:

    .. math::

        \text{Min} =
        \begin{cases}
            f(8.05502, 9.66459)   &= -19.2085 \\
            f(-8.05502, 9.66459)  &= -19.2085 \\
            f(8.05502, -9.66459)  &= -19.2085 \\
            f(-8.05502, -9.66459) &= -19.2085
        \end{cases}

    Search domain:

    .. math::

        -10 \leq x_1, x_2 \leq 10

    **References**: *Test functions for optimization* (Wikipedia):
    https://en.wikipedia.org/wiki/Test_functions_for_optimization.

    .. image:: holder_3d.png

    .. image:: holder.png

    Example
    -------

    To evaluate a single 2D point :math:`x = \begin{pmatrix} 0 \\ 0 \end{pmatrix}`:

    >>> holder( np.array([0, 0]) )
    0.0

    The result should be :math:`f(x) = 0`.

    Example
    -------

    To evaluate multiple 2D points :math:`x_1 = \begin{pmatrix} 0 \\ 0 \end{pmatrix}`,
    :math:`x_2 = \begin{pmatrix} 0 \\ 1 \end{pmatrix}` and
    :math:`x_3 = \begin{pmatrix} 1 \\ 0 \end{pmatrix}` at once:

    >>> holder( np.array([[0., 0., 1.], [0., 1., 0.]]) )
    ... # doctest: +NORMALIZE_WHITESPACE
    array([-0. , -0. , -1.66377043])

    The result should be :math:`f(x_1) = 0`, :math:`f(x_2) = 0` and :math:`f(x_3) = -1.66377043`.

    Parameters
    ----------
    x : array_like
        One dimension Numpy array of the point at which the Hölder table function is to be computed
        or a two dimension Numpy array of points at which the Hölder table function is to be computed.

    Returns
    -------
    float or array_like
        The value(s) of the Hölder table function for the given point(s) `x`.
    """
    assert x.shape[0] == 2, x.shape
    return -np.abs(np.sin(x[0]) * np.cos(x[1]) * np.exp(np.abs(1.0 - np.sqrt(x[0]**2.0 + x[1]**2.0)/np.pi )))


class Holder(_ObjectiveFunction):
    """
    TODO
    """
    def __init__(self, ndim):
        super().__init__()

        self._objective_function = holder

        self.ndim = ndim
        if self.ndim != 2:
            raise ValueError("The holder function is defined for solution spaces having 2 dimensions.")

        self.bounds = np.ones((2, self.ndim))    # TODO: take this or the transpose of this ?
        self.bounds[0,:] = -10.
        self.bounds[1,:] =  10.

        self.continuous = True

    @property
    def unimodal(self):
        return False

    @property
    def arg_min(self) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        # https://en.wikipedia.org/wiki/Test_functions_for_optimization
        return (np.array([8.05502, 9.66459]), np.array([-8.05502, 9.66459]), np.array([8.05502, -9.66459]), np.array([-8.05502, -9.66459]))


holder2d = Holder(ndim=2)


# -*- coding: utf-8 -*-

# Copyright (c) 2013,2014,2015,2016,2017,2024 Jeremie DECOCK (http://www.jdhp.org)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#  
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from dataclasses import dataclass
import numpy as np
from typing import Callable

from .optimizer import Optimizer


@dataclass
class GeneticAlgorithmResult:
    best_solution: np.ndarray
    best_objective: float
    solution_history: np.ndarray
    objective_history: np.ndarray
    best_solution_history: np.ndarray
    best_objective_history: np.ndarray


class GeneticAlgorithm(Optimizer):

    def minimize(
        self,
        objective_function: Callable[[np.ndarray], float],
        init_pop_mean: np.ndarray,
        init_pop_std: np.ndarray,
        num_gen: int = 50,
        pop_size: int = 100,
        mutation_rate: float = 0.01,
        plot: bool = False
    ) -> GeneticAlgorithmResult:
        """
        Minimize the objective function using a genetic algorithm.
    
        Parameters
        ----------
        objective_function : Callable[[np.ndarray], float]
            The objective function to minimize.
        init_pop_mean : np.ndarray
            The mean of the initial population. Shape: (n_dimensions,)
        init_pop_std : np.ndarray
            The standard deviation of the initial population. Shape: (n_dimensions,)
        num_gen : int, optional
            The number of generations (iterations) to run the algorithm (default is 50).
        pop_size : int, optional
            The size of the population (default is 100).
        mutation_rate : float, optional
            The mutation rate (default is 0.01).
        plot : bool, optional
            Whether to plot the progress of the algorithm (default is False).
    
        Returns
        -------
        best_solution : np.ndarray
            The best solution found by the algorithm. Shape: (n_dimensions,)
        best_objective : float
            The objective value of the best solution.
        solution_history : np.ndarray
            The history of solutions over generations. Shape: (n_dimensions, num_gen, pop_size)
        objective_history : np.ndarray
            The history of objective values over generations. Shape: (1, num_gen, pop_size)
        """
        
        # TODO vérif cohérence pop_size et init_pop_mean / init_pop_std
    
        # Solution history is a 3D array:
        #   - 1st dimension: the solutions dimension
        #   - 2nd dimension: the generation number (or iteration number)
        #   - 3rd dimension: the individuals dimension
        # Thus with a run of 100 iterations, a 2D objective function and a population size of 30, solution_history will have a shape of (2, 100, 30)
        solution_history = np.zeros((len(init_pop_mean), num_gen, pop_size))
    
        # Objective history is a 3D array:
        #   - 1st dimension: the number of objectives to optimize (by default 1)
        #   - 2nd dimension: the generation number (or iteration number)
        #   - 3rd dimension: the individuals dimension
        # Thus with a run of 100 iterations, a mono objective function and a population size of 30, solution_history will have a shape of (1, 100, 30)
        objective_history = np.zeros((1, num_gen, pop_size))
    
        # Initialisation de la population
        population = np.random.normal(init_pop_mean, init_pop_std, (pop_size, len(init_pop_mean)))
    
        for generation in range(num_gen):
            # Évaluation
            fitness = np.array([objective_function(ind) for ind in population])
    
            # Enregistrement de la population et des valeurs d'objectifs
            solution_history[:, generation, :] = population.T
            objective_history[0, generation, :] = fitness
    
            # Sélection
            selected_indices = np.argsort(fitness)[:pop_size // 2]
            selected_population = population[selected_indices]
    
            # Croisement
            new_population = []
            for _ in range(pop_size // 2):
                parents = selected_population[np.random.choice(len(selected_population), 2, replace=False)]
                if len(parents[0]) > 2:
                    crossover_point = np.random.randint(1, len(parents[0]) - 1)
                else:
                    crossover_point = 1  # or some other logic
                child1 = np.concatenate((parents[0][:crossover_point], parents[1][crossover_point:]))
                child2 = np.concatenate((parents[1][:crossover_point], parents[0][crossover_point:]))
                new_population.extend([child1, child2])
    
            # Mutation
            new_population = np.array(new_population)
            mutations = np.random.rand(*new_population.shape) < mutation_rate
            new_population = new_population + mutations * np.random.normal(0, 1, new_population.shape)
    
            population = new_population
    
        # Retourner le meilleur individu
        best_index = np.argmin([objective_function(ind) for ind in population])
        best_solution = population[best_index]
        best_objective = objective_function(best_solution)

        # Historique du meilleur individu pour chaque génération
        best_objective_history = objective_history.argmin(axis=2)
        best_solution_history = solution_history[:, :, best_objective_history]

        # best_solution_history = solution_history[:, :, best_index]
        # best_objective_history = objective_history[0, :, best_index]

        # Historique du meilleur individu pour chaque génération
        best_solution_index_history = objective_history.argmin(axis=2).flatten()
        best_objective_history = objective_history[0, np.arange(num_gen), best_solution_index_history]
        best_solution_history = solution_history[:, np.arange(num_gen), best_solution_index_history]

        result = GeneticAlgorithmResult(
            best_solution=best_solution,
            best_objective=best_objective,
            solution_history=solution_history,
            objective_history=objective_history,
            best_solution_history=best_solution_history,
            best_objective_history=best_objective_history
        )
    
        return result



# -*- coding: utf-8 -*-

# Copyright (c) 2013,2014,2015,2016,2017,2024 Jeremie DECOCK (http://www.jdhp.org)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#  
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from dataclasses import dataclass
import numpy as np

from .optimizer import Optimizer


@dataclass
class CEMResult:
    best_solution: np.ndarray
    best_objective: float
    solution_history: np.ndarray
    objective_history: np.ndarray


class CEM(Optimizer):

    def minimize(self,
                 objective_function,
                 init_solution,
                 standard_deviation=1.,
                 num_samples=100,
                 elite_frac=0.2,
                 num_iterations=50,
                 plot=False):

        n_elite = int(num_samples * elite_frac)
        mean = init_solution
        std = np.ones_like(init_solution) * standard_deviation

        solution_history = []
        objective_history = []

        for i in range(num_iterations):
            samples = np.random.normal(loc=mean, scale=std, size=(num_samples, len(init_solution)))
            objectives = np.array([objective_function(sample) for sample in samples])

            elite_indices = objectives.argsort()[:n_elite]
            elite_samples = samples[elite_indices]

            mean = np.mean(elite_samples, axis=0)
            std = np.std(elite_samples, axis=0)

            best_solution = mean
            best_objective = objective_function(best_solution)

            solution_history.append(best_solution)
            objective_history.append(best_objective)

        solution_history = np.array(solution_history).T
        objective_history = np.array(objective_history)

        result = CEMResult(
            best_solution=best_solution,
            best_objective=best_objective,
            solution_history=solution_history,
            objective_history=objective_history
        )

        return result


# -*- coding: utf-8 -*-

# Copyright (c) 2013,2014,2015,2016,2017 Jeremie DECOCK (http://www.jdhp.org)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#  
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import numpy as np
import warnings

class Optimizer(object):
    """
    Optimizer class.
    By default, all optimizers works in minimization.
    """

    optimizer_name = "unknown"

    def __init__(self):
        self.log = Log()

    def plotSamples(self, x_hist_array, y_hist_array, nabla_hist_array=None, objective_function=None, save_filename=None):
        """
        Plot the objective function for x_hist_array and the evaluated points.
        This only works for 1D and 2D functions.
        """
        import matplotlib.pyplot as plt

        assert x_hist_array.ndim == 2, x_hist_array.ndim
        assert y_hist_array.ndim == 1, y_hist_array.ndim
        assert y_hist_array.shape[0] == x_hist_array.shape[0], y_hist_array.shape

        if x_hist_array.shape[1]==1:
            # 1D case

            fig = plt.figure(figsize=(16.0, 10.0))
            ax = fig.add_subplot(111)

            # PLOT THE OBJECTIVE FUNCTION 

            if objective_function is not None:
                # BUILD DATA

                assert objective_function.domain_min.ndim == 1
                assert objective_function.domain_max.ndim == 1
                assert objective_function.domain_min.shape[0] == 1
                assert objective_function.domain_max.shape[0] == 1

                xmin = objective_function.domain_min[0]
                xmax = objective_function.domain_max[0]
                assert xmin < xmax

                xstep = (xmax - xmin) / 1000.

                x_vec = np.arange(xmin, xmax, xstep)
                y_vec = objective_function(x_vec.reshape([-1, 1]))
                ax.plot(x_vec, y_vec, "-", label="objective function")

            # PLOT VISITED POINTS
            ax.plot(x_hist_array[:,0], y_hist_array, ".", label="visited points")
            
            # PLOT THE BEST VISITED POINTS
            x_min = x_hist_array[y_hist_array.argmin(), :]
            y_min = y_hist_array.min()
            ax.plot(x_min, y_min, "xr")

            # PLOT GRADIENT OF VISITED POINTS
            if nabla_hist_array is not None:
                pass

            # TITLE AND LABELS
            ax.set_title('Visited points', fontsize=20)
            ax.set_xlabel(r"$x$", fontsize=32)
            ax.set_ylabel(r"$f(x)$", fontsize=32)

            # LEGEND
            ax.legend(loc='lower right', fontsize=20)

            # SAVE FILES ######################
            if save_filename is not None:
                filename = save_filename + ".pdf"
                plt.savefig(filename)

            # PLOT
            plt.show()

        elif x_hist_array.shape[1]==2:
            # 2D case

            from mpl_toolkits.mplot3d import axes3d
            if objective_function is not None:
                from mpl_toolkits.mplot3d import axes3d
                from matplotlib import cm

            fig = plt.figure()

            if objective_function is not None:
                ax = fig.gca(projection='3d')
            else:
                ax = axes3d.Axes3D(fig)

            # PLOT THE OBJECTIVE FUNCTION 

            if objective_function is not None:
                # BUILD DATA

                assert objective_function.domain_min.ndim == 1
                assert objective_function.domain_max.ndim == 1
                assert objective_function.domain_min.shape[0] == 2
                assert objective_function.domain_max.shape[0] == 2

                x1min = objective_function.domain_min[0]
                x1max = objective_function.domain_max[0]
                assert x1min < x1max

                x2min = objective_function.domain_min[1]
                x2max = objective_function.domain_max[1]
                assert x2min < x2max

                x1step = (x1max - x1min) / 200.
                x2step = (x2max - x2min) / 200.

                range_x1 = np.arange(x1min, x1max, x1step)
                range_x2 = np.arange(x2min, x2max, x2step)

                mesh_x1,mesh_x2 = np.meshgrid(range_x1, range_x2)

                # TODO: take advantage of meshgrid, for now, it's not optimized at
                #       all and not very well written
                z = np.zeros(mesh_x1.shape)         
                for x1i in range(z.shape[0]):
                    for x2i in range(z.shape[1]):
                        point = np.array([mesh_x1[x1i, x2i], mesh_x2[x1i, x2i]])
                        z[x1i, x2i] = objective_function(point)

                # PLOT
                ax.plot_surface(mesh_x1, mesh_x2, z, rstride=5, cstride=5, alpha=0.3)
                cset = ax.contourf(mesh_x1, mesh_x2, z, zdir='z', offset=0, cmap=cm.coolwarm)

            # PLOT VISITED POINTS
            ax.scatter(x_hist_array[:,0], x_hist_array[:,1], y_hist_array, color='b')
            
            # PLOT THE BEST VISITED POINT
            x_min = x_hist_array[y_hist_array.argmin(), :]
            y_min = y_hist_array.min()
            ax.scatter(x_min[0], x_min[1],  y_min, color='r')

            # PLOT GRADIENT OF VISITED POINTS
            if nabla_hist_array is not None:
                pass

            # TITLE AND LABELS
            ax.set_title('Visited points', fontsize=20)
            ax.set_xlabel(r'$x_1$', fontsize=32)
            ax.set_ylabel(r'$x_2$', fontsize=32)
            ax.set_zlabel(r'$f(x)$', fontsize=32)

            # SAVE FILES ######################
            if save_filename is not None:
                filename = save_filename + ".pdf"
                plt.savefig(filename)

            plt.show()

        else:
            warnings.warn("Cannot plot samples: too many dimensions.")


    def plotCosts(self, y_array):
        """
        Plot the evolution of point's cost evaluated during iterations.
        y_array[i] is the cost at the ith iteration.
        y_array must be a numpy array of dimension 1.
        """
        import matplotlib.pyplot as plt

        # y_array must be a numpy array of dimension 1
        assert y_array.ndim == 1, "y_array.ndim = " + str(y_array.ndim)

        fig = plt.figure(figsize=(16.0, 10.0))
        ax = fig.add_subplot(111)
        ax.plot(y_array, "-", label="value")

        # TITLE AND LABELS
        ax.set_title("Value over iterations", fontsize=20)
        ax.set_xlabel(r"iteration $i$", fontsize=32)
        ax.set_ylabel(r"$f(x)$", fontsize=32)

        # LEGEND
        ax.legend(loc='lower right', fontsize=20)

        # PLOT
        plt.show()


class Log:
    # TODO: this class is not used yet ?
    def __init__(self):
        self.data = {}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (c) 2017 Jeremie DECOCK (http://www.jdhp.org)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
TODO
"""

__all__ = ['GaussianNoise', 'additive_gaussian_noise', 'multiplicative_gaussian_noise',
           'PoissonNoise', 'additive_poisson_noise', 'multiplicative_poisson_noise',
           'UniformNoise', 'additive_uniform_noise', 'multiplicative_uniform_noise']

import numpy as np

class GaussianNoise:
    """Noise class for objective functions.

    TODO
    """
    def __init__(self, loc=0., scale=1., noise_type='additive'):
        self.noise_type = noise_type
        self.loc = loc
        self.scale = scale

    def __call__(self, x, y):
        """
        TODO
        """
        rvs = np.random.normal(loc=self.loc, scale=self.scale, size=y.shape)

        if self.noise_type == 'additive':
            y += rvs
        elif self.noise_type == 'multiplicative':
            y += y * rvs
        else:
            raise ValueError("Unknown value {}.".format(self.noise_type))

        return y

additive_gaussian_noise = GaussianNoise(loc=0., scale=1., noise_type='additive')
multiplicative_gaussian_noise = GaussianNoise(loc=0., scale=0.1, noise_type='multiplicative')


class PoissonNoise:
    """Noise class for objective functions.

    TODO
    """
    def __init__(self, lam=1., noise_type='additive'):
        self.noise_type = noise_type
        self.lam = lam

    def __call__(self, x, y):
        """
        TODO
        """
        rvs = np.random.poisson(lam=self.lam, size=y.shape)

        if self.noise_type == 'additive':
            y += rvs
        elif self.noise_type == 'multiplicative':
            y += y * rvs
        else:
            raise ValueError("Unknown value {}.".format(self.noise_type))

        return y

additive_poisson_noise = PoissonNoise(lam=3., noise_type='additive')
multiplicative_poisson_noise = PoissonNoise(lam=3., noise_type='multiplicative')


class UniformNoise:
    """Noise class for objective functions.

    TODO
    """
    def __init__(self, high=0., low=1., noise_type='additive'):
        self.noise_type = noise_type
        self.high = high
        self.low = low

    def __call__(self, x, y):
        """
        TODO
        """
        rvs = np.random.uniform(high=self.high, low=self.low, size=y.shape)

        if self.noise_type == 'additive':
            y += rvs
        elif self.noise_type == 'multiplicative':
            y += y * rvs
        else:
            raise ValueError("Unknown value {}.".format(self.noise_type))

        return y

additive_uniform_noise = UniformNoise(high=0., low=1., noise_type='additive')
multiplicative_uniform_noise = UniformNoise(high=0., low=1., noise_type='multiplicative')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (c) 2017 Jeremie DECOCK (http://www.jdhp.org)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
TODO

**TODO**:

Plots:
- [x] visited solutions in the solution space (2D plots with pcolormesh and contour)
- [x] f(x*) w.t. iteration number (well... actually this one is pretty useless...)
- [x] error w.t. iteration number (=> + fit to have a clear view of the convergence rate ?)
- [x] error w.t. number of function evaluations + error w.t. *total* number of function evaluations (i.e. including the number of gradient and hessian evaluations)
- [x] error w.t. execution time
- [ ] error w.t. ... => add an option to choose between the current solution or the best current solution 
- [ ] (benchmark session ! distinguish the derivative-free to the non-derivative free case) average version of 3., 4., 5. over several runs with random initial state (+ error bar or box plot)
- [ ] (benchmark session) err w.t. algorithms parameters (plot the iteration or evaluation number or execution time to reach in average an error lower than N% with e.g. N=99%)
"""

__all__ = ['array_list_to_array',
           'plot_contour_2d_solution_space',
           'plot_2d_solution_space',
           'plot_fx_wt_iteration_number',
           'plot_err_wt_iteration_number',
           'plot_err_wt_execution_time',
           'plot_err_wt_num_feval']

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.colors as colors
from mpl_toolkits.mplot3d import Axes3D
from typing import Callable, Dict, List, Optional, Tuple, Union

def array_list_to_array(y_array_list):
    """Convert a sequence of 1D arrays (with possibly different sizes) to a 2D array.

    When given 1D arrays don't have the same size, "missing" items at the end
    of each 1D array are replaced by np.nan.

    Parameters
    ----------
    y_array_list : sequence of ndarray
        The sequence of 1D numpy arrays (with possibly different sizes) to
        join.

    Returns
    -------
    ndarray
        The 2D numpy array made from the concatenation `y_array_list` items.
    """
    max_size = max((y_array.shape[0] for y_array in y_array_list))
    y_array = np.full((len(y_array_list), max_size), np.nan)

    for index, y_array_i in enumerate(y_array_list):
        y_array[index,:y_array_i.shape[0]] = y_array_i

    return y_array


def plot_1d(y,
            x=None,
            ax=None,
            x_label=None,
            y_label=None,
            legend=None,
            title=None,
            x_log=False,
            y_log=False,
            plot_option=None,
            **kwargs):
    """Make a 1D plot.

    Parameters
    ----------
    y : ndarray
        The 1D or 2D numpy array containing the data's y-coordinate (ordinate)
        to plot. If `y` is a 2D array, it is aggregated *on the first dimension*
        (e.g. `y = np.nanmean(y, axis=0)`) which means the `y` array should be
        defined like the following:
        `y = np.array([sample_1, sample_2, ..., sample_n])`
        where `sample_i` is a 1D numpy array containing for instance the error
        of (the ith execution of) a minimizer w.t. the evaluation number
        (i.e. `sample_1` is the errors of a first execution of the algorithm,
        `sample_2` is the errors of a second execution of the algorithm,
        ...).
    x : ndarray
        The 1D numpy array containing the data's x-coordinate (abscissa) to
        plot.
    ax : AxesSubplot
        The Matplotlib axis object to plot on. If `ax` is `None`, a new axis is
        created and automatically shown. Otherwise, the `ax` object is updated
        with the plot but is not automatically shown.
    x_label : string
        The label for the x-axis.
    y_label : string
        The label for the y-axis.
    legend : string
        The legend label.
    title : string
        The plot title.
    x_log : bool
        If `True`, a log scale is used on the x-axis.
    y_log : bool
        If `True`, a log scale is used on the y-axis.
    plot_option : string
        Define the aggregation method if `y` is a 2D array. Possible values are
        "mean" or "median".
    **kwargs
        Any arbitrary keyword arguments accepted by `matplotlib.pyplot.plot()`.
    """
    if y.ndim == 2:
        # Aggregate data
        if plot_option == 'mean':
            y = np.nanmean(y, axis=0)
        elif plot_option == 'median':
            y = np.nanmedian(y, axis=0)
        else:
            raise ValueError("Two dimension arrays require a `plot_option` to define the aggregation method.")
    elif y.ndim != 1:
        raise ValueError("Wrong number of dimensions: {}.".format(y.ndim))

    if ax is None:
        fig, ax = plt.subplots(figsize=(12, 8))

    if x_log:
        ax.set_xscale("log", nonposx='clip')   # TODO

    if y_log:
        ax.set_yscale("log", nonposy='clip')   # TODO

    if x is None:
        ax.plot(y, **kwargs)
    else:
        ax.plot(x, y, **kwargs)

    if legend is not None:
        ax.legend(loc='best')

    if x_label is not None:
        ax.set_xlabel(x_label)

    if y_label is not None:
        ax.set_ylabel(y_label)

    if title is not None:
        ax.set_title(title)

    if ax is None:
        plt.show()


def plot_fx_wt_iteration_number(fx_array, **kwargs):
    """Plot f(x*) w.t. iteration number.

    TODO
    """
    plot_1d(y=fx_array,
            x_label="Iteration number",
            y_label="f(x)",
            **kwargs)


def plot_err_wt_iteration_number(error_array, **kwargs):
    """Plot error w.t. iteration number.

    TODO
    """
    plot_1d(y=error_array,
            x_label="Iteration number",
            y_label="Error",
            **kwargs)


def plot_err_wt_execution_time(error_array, time_array, **kwargs):
    """Plot error w.t. execution time.

    TODO
    """
    plot_1d(y=error_array,
            x=time_array,
            x_label="Time (sec)",
            y_label="Error",
            **kwargs)


def plot_err_wt_num_feval(error_array, num_eval_array=None, **kwargs):
    """Plot error w.t. number of function evaluations.

    TODO
    """
    plot_1d(y=error_array,
            x=num_eval_array,
            x_label="Num fonction evaluations",
            y_label="Error",
            **kwargs)





# def plot_2d_solution_space(func,
#                            fig=None,
#                            ax=None,
#                            show=True,
#                            xmin=-np.ones(2),
#                            xmax=np.ones(2),
#                            xstar=None,
#                            xvisited=None,
#                            angle_view=None,
#                            x1_grid_size=100,
#                            x2_grid_size=100,
#                            figsize=(12, 8),
#                            x1_label=r"$x_1$",
#                            x2_label=r"$x_2$",
#                            zlabel=r"$f(x_1, x_2)$",
#                            xstar_label=r"$x^*$",
#                            title_fontsize=12,
#                            label_fontsize=12,
#                            legend_fontsize=12,
#                            title=""):
#     """Plot points visited during the execution of an optimization algorithm.

#     TODO
#     """
#     if fig is None or ax is None:                # TODO
#         fig = plt.figure(figsize=figsize)
#         ax = axes3d.Axes3D(fig)

#     if angle_view is not None:
#         ax.view_init(angle_view[0], angle_view[1])

#     x1_space = np.linspace(xmin[0], xmax[0], x1_grid_size)
#     x2_space = np.linspace(xmin[1], xmax[1], x2_grid_size)

#     x1_mesh, x2_mesh = np.meshgrid(x1_space, x2_space)

#     zz = func(np.array([x1_mesh.ravel(), x2_mesh.ravel()])).reshape(x1_mesh.shape)   # TODO

#     ############################

#     surf = ax.plot_surface(x1_mesh,
#                            x2_mesh,
#                            zz,
#                            cmap='gnuplot2', # 'jet' # 'gnuplot2'
#                            norm=colors.LogNorm(),   # TODO
#                            rstride=1,
#                            cstride=1,
#                            #color='b',
#                            shade=False)

#     ax.set_zlabel(zlabel, fontsize=label_fontsize)

#     fig.colorbar(surf, shrink=0.5, aspect=5)

#     ############################

#     if xstar is not None:
#         ax.scatter(xstar[0],
#                    xstar[1],
#                    func(xstar),
#                    #s=50,          # TODO
#                    c='red',
#                    alpha=1,
#                    label=xstar_label)

#     ax.set_title(title, fontsize=title_fontsize)

#     ax.set_xlabel(x1_label, fontsize=label_fontsize)
#     ax.set_ylabel(x2_label, fontsize=label_fontsize)

#     ax.legend(fontsize=legend_fontsize)

#     if show:
#         plt.show()

#     return fig, ax


###############################################################################

def plot_2d_solution_space(func,
                           xmin=-np.ones(2),
                           xmax=np.ones(2),
                           xstar=None,
                           angle_view=None,
                           title="",
                           vmin=None,
                           vmax=None,
                           zlog=True,
                           vectorized_func=True,
                           output_file_name=None):
    """TODO
    """
    fig = plt.figure(figsize=(12, 8))
    ax = Axes3D(fig)
    
    if angle_view is not None:
        ax.view_init(angle_view[0], angle_view[1])

    x1_space = np.linspace(xmin[0], xmax[0], 100)
    x2_space = np.linspace(xmin[1], xmax[1], 100)
    
    x1_mesh, x2_mesh = np.meshgrid(x1_space, x2_space)

    if vectorized_func:
        zz = vectorized_grid_eval(func, x1_mesh, x2_mesh)
    else:
        zz = grid_eval(func, x1_mesh, x2_mesh)

    ############################
    
    if xstar.ndim == 1:
        min_value = func(xstar)
    else:
        min_value = min(func(xstar))
    max_value = zz.max()
    
    if vmin is None:
        if zlog:
            vmin = 0.1             # TODO
        else:
            vmin = min_value
        
    if vmax is None:
        vmax = max_value

    if zlog:
        norm = colors.LogNorm(vmin=vmin, vmax=vmax)
    else:
        norm = colors.Normalize(vmin=vmin, vmax=vmax)
        
    surf = ax.plot_surface(x1_mesh,
                           x2_mesh,
                           zz,
                           cmap='gnuplot2', # 'jet' # 'gnuplot2'
                           norm=norm,
                           rstride=1,
                           cstride=1,
                           #color='b',
                           shade=False)

    ax.set_zlabel(r"$f(x_1, x_2)$")

    fig.colorbar(surf, shrink=0.5, aspect=5)


###############################################################################

def grid_eval(func, x1_mesh, x2_mesh):
    
    z = [func([x, y]) for x, y in zip(x1_mesh.ravel(), x2_mesh.ravel())]
    zz = np.array(z).reshape(x1_mesh.shape)

    return zz


def vectorized_grid_eval(func, x1_mesh, x2_mesh):
    
    zz = func(np.array([x1_mesh.ravel(), x2_mesh.ravel()])).reshape(x1_mesh.shape)

    return zz

###############################################################################


def plot_2d_contour_solution_space(
    func,
    xmin=-np.ones(2),
    xmax=np.ones(2),
    xstar: Union[np.ndarray, Tuple[np.ndarray, ...]] = None,
    title="",
    vmin=None,
    vmax=None,
    zlog=True,
    output_file_name=None,
    vectorized_func=True,
    ax=None
):
    """TODO
    """
    if ax is None:
        fig, ax = plt.subplots(figsize=(12, 8))

    x1_space = np.linspace(xmin[0], xmax[0], 200)
    x2_space = np.linspace(xmin[1], xmax[1], 200)
    
    x1_mesh, x2_mesh = np.meshgrid(x1_space, x2_space)

    if vectorized_func:
        zz = vectorized_grid_eval(func, x1_mesh, x2_mesh)
    else:
        zz = grid_eval(func, x1_mesh, x2_mesh)
    
    ############################
    
    if isinstance(xstar, tuple):
        # Multimodal objective function
        xs = xstar[0]
    else:
        # Monomodal objective function
        xs = xstar

    if xs.ndim == 1:
        min_value = func(xs)
    else:
        min_value = min(func(xs))
    max_value = zz.max()
    
    if vmin is None:
        if zlog:
            vmin = 0.1             # TODO
        else:
            vmin = min_value
        
    if vmax is None:
        vmax = max_value

    if zlog:
        norm = colors.LogNorm(vmin=vmin, vmax=vmax)
    else:
        norm = colors.Normalize(vmin=vmin, vmax=vmax)
    
    levels = np.logspace(0.1, 3., 5)          # TODO

    im = ax.pcolormesh(x1_mesh, x2_mesh, zz,
                       norm=norm,
                       shading='gouraud',
                       cmap='gnuplot2') # 'jet' # 'gnuplot2'

    plt.colorbar(im, ax=ax)

    cs = plt.contour(x1_mesh, x2_mesh, zz, levels,
                     linewidths=(2, 2, 2, 2, 3),
                     linestyles=('dotted', '-.', 'dashed', 'solid', 'solid'),
                     alpha=0.5,
                     colors='white')
    ax.clabel(cs, inline=False, fontsize=12)

    ############################

    if xstar is not None:
        if isinstance(xstar, tuple):
            # Multimodal objective function
            for _xstar in xstar:
                ax.scatter(
                    _xstar[0],
                    _xstar[1],
                    c='red',
                    label="$x^*$"
                )
        else:
            ax.scatter(
                xstar[0],
                xstar[1],
                c='red',
                label="$x^*$"
            )

    ax.set_title(title)

    ax.set_xlabel(r"$x_1$")
    ax.set_ylabel(r"$x_2$")

    ax.legend(fontsize=12)
    
    if output_file_name is not None:
        plt.savefig(output_file_name, transparent=True)

    # plt.show()

    return ax


def plot_contour_2d_solution_space(func,
                                   fig=None,
                                   ax=None,
                                   show=True,
                                   xmin=-np.ones(2),
                                   xmax=np.ones(2),
                                   xstar=None,
                                   xvisited=None,
                                   x1_grid_size=200,
                                   x2_grid_size=200,
                                   figsize=(12, 8),
                                   x1_label=r"$x_1$",
                                   x2_label=r"$x_2$",
                                   xstar_label=r"$x^*$",
                                   title_fontsize=12,
                                   label_fontsize=12,
                                   legend_fontsize=12,
                                   title=""):
    """Plot points visited during the execution of an optimization algorithm.

    TODO
    """
    if (fig is None) or (ax is None):                # TODO
        fig, ax = plt.subplots(figsize=figsize)

    if xvisited is not None:
        xmin = np.amin(np.hstack([xmin.reshape([-1, 1]), xvisited]), axis=1)
        xmax = np.amax(np.hstack([xmax.reshape([-1, 1]), xvisited]), axis=1)

    x1_space = np.linspace(xmin[0], xmax[0], x1_grid_size)
    x2_space = np.linspace(xmin[1], xmax[1], x2_grid_size)

    x1_mesh, x2_mesh = np.meshgrid(x1_space, x2_space)

    zz = func(np.array([x1_mesh.ravel(), x2_mesh.ravel()])).reshape(x1_mesh.shape)

    ############################

    if xstar is not None:
        min_value = func(xstar)
    else:
        min_value = zz.min()

    max_value = zz.max()

    levels = np.logspace(0.1, 3., 5)          # TODO

    im = ax.pcolormesh(x1_mesh, x2_mesh, zz,
                       vmin=0.1,              # TODO
                       vmax=max_value,
                       norm=colors.LogNorm(), # TODO
                       shading='gouraud',
                       cmap='gnuplot2') # 'jet' # 'gnuplot2'

    plt.colorbar(im, ax=ax)

    cs = plt.contour(x1_mesh, x2_mesh, zz, levels,
                     linewidths=(2, 2, 2, 2, 3),
                     linestyles=('dotted', '-.', 'dashed', 'solid', 'solid'),
                     alpha=0.5,
                     colors='white')
    ax.clabel(cs, inline=False, fontsize=label_fontsize)

    ############################

    if xvisited is not None:
        ax.plot(xvisited[0],
                xvisited[1],
                '-og',
                alpha=0.5,
                label="$visited$")

    ############################

    if xstar is not None:
        sc = ax.scatter(xstar[0],
                   xstar[1],
                   c='red',
                   label=xstar_label)
        sc.set_zorder(10)        # put this point above every thing else

    ############################

    ax.set_title(title, fontsize=title_fontsize)

    ax.set_xlabel(x1_label, fontsize=label_fontsize)
    ax.set_ylabel(x2_label, fontsize=label_fontsize)

    ax.legend(fontsize=legend_fontsize)

    if show:
        plt.show()

    return fig, ax


# -*- coding: utf-8 -*-

# Copyright (c) 2013,2014,2015,2016,2017,2024 Jeremie DECOCK (http://www.jdhp.org)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#  
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from dataclasses import dataclass
import numpy as np

from .optimizer import Optimizer


@dataclass
class CMAESResult:
    best_solution: np.ndarray
    best_objective: float
    solution_history: np.ndarray
    objective_history: np.ndarray


class CMAES(Optimizer):

    def minimize(
        self,
        objective_function,
        init_solution,
        standard_deviation=1.,
        num_iterations=50,
        population_size=10,
        plot=False
    ) -> CMAESResult:
        n = len(init_solution)
        mean = np.array(init_solution)
        sigma = standard_deviation
        population_size = population_size
        weights = np.log(population_size + 0.5) - np.log(np.arange(1, population_size + 1))
        weights /= np.sum(weights)
        mu_eff = 1 / np.sum(weights ** 2)
        c_sigma = (mu_eff + 2) / (n + mu_eff + 5)
        d_sigma = 1 + 2 * max(0, np.sqrt((mu_eff - 1) / (n + 1)) - 1) + c_sigma
        c_c = (4 + mu_eff / n) / (n + 4 + 2 * mu_eff / n)
        c_1 = 2 / ((n + 1.3) ** 2 + mu_eff)
        c_mu = min(1 - c_1, 2 * (mu_eff - 2 + 1 / mu_eff) / ((n + 2) ** 2 + mu_eff))
        p_c = np.zeros(n)
        p_sigma = np.zeros(n)
        B = np.eye(n)
        D = np.ones(n)
        C = B @ np.diag(D ** 2) @ B.T
        inv_sqrt_C = B @ np.diag(D ** -1) @ B.T
        best_solution = mean
        best_objective = objective_function(best_solution)

        solution_history = []
        objective_history = []

        for i in range(num_iterations):
            population = np.random.multivariate_normal(mean, sigma ** 2 * C, population_size)
            fitness = np.array([objective_function(ind) for ind in population])
            indices = np.argsort(fitness)
            population = population[indices]
            fitness = fitness[indices]

            mean = np.dot(weights, population[:len(weights)])
            p_sigma = (1 - c_sigma) * p_sigma + np.sqrt(c_sigma * (2 - c_sigma) * mu_eff) * inv_sqrt_C @ (mean - best_solution) / sigma
            h_sigma = np.linalg.norm(p_sigma) / np.sqrt(1 - (1 - c_sigma) ** (2 * (i + 1))) < (1.4 + 2 / (n + 1)) * np.sqrt(n)
            p_c = (1 - c_c) * p_c + h_sigma * np.sqrt(c_c * (2 - c_c) * mu_eff) * (mean - best_solution) / sigma
            artmp = (population[:len(weights)] - best_solution) / sigma
            C = (1 - c_1 - c_mu) * C + c_1 * (np.outer(p_c, p_c) + (1 - h_sigma) * c_c * (2 - c_c) * C) + c_mu * artmp.T @ np.diag(weights) @ artmp
            sigma *= np.exp((c_sigma / d_sigma) * (np.linalg.norm(p_sigma) / np.sqrt(1 - (1 - c_sigma) ** (2 * (i + 1))) - 1))

            best_solution = population[0]
            best_objective = fitness[0]

            solution_history.append(best_solution)
            objective_history.append(best_objective)

        solution_history = np.array(solution_history).T
        objective_history = np.array(objective_history)

        result = CMAESResult(
            best_solution=best_solution,
            best_objective=best_objective,
            solution_history=solution_history,
            objective_history=objective_history
        )

        return result


#!/usr/bin/env python3

# import logging

# logging.basicConfig(
#     level=logging.INFO,
#     format='%(asctime)s %(levelname)s:%(name)s:%(message)s',  # See https://docs.python.org/3/library/logging.html#logrecord-attributes
#     datefmt=r'%Y-%m-%dT%H:%M:%S%z'
# )

def main():
    pass

if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

# Copyright (c) 2013,2014,2015,2016,2017,2024 Jeremie DECOCK (http://www.jdhp.org)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#  
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from dataclasses import dataclass
import numpy as np
from typing import Callable

from .optimizer import Optimizer

@dataclass
class PSOResult:
    best_solution: np.ndarray
    best_objective: float
    solution_history: np.ndarray
    objective_history: np.ndarray
    best_solution_history: np.ndarray
    best_objective_history: np.ndarray


class ParticleSwarmOptimizer(Optimizer):

    def minimize(
        self,
        objective_function: Callable[[np.ndarray], float],
        init_pop_mean: np.ndarray,
        init_pop_std: np.ndarray,
        num_gen: int = 50,
        pop_size: int = 100,
        inertia: float = 0.5,
        cognitive: float = 1.5,
        social: float = 1.5,
        plot: bool = False
    ) -> PSOResult:
        """
        Minimize the objective function using a particle swarm optimization algorithm.
    
        Parameters
        ----------
        objective_function : Callable[[np.ndarray], float]
            The objective function to minimize.
        init_pop_mean : np.ndarray
            The mean of the initial population. Shape: (n_dimensions,)
        init_pop_std : np.ndarray
            The standard deviation of the initial population. Shape: (n_dimensions,)
        num_gen : int, optional
            The number of generations (iterations) to run the algorithm (default is 50).
        pop_size : int, optional
            The size of the population (default is 100).
        inertia : float, optional
            The inertia coefficient (default is 0.5).
        cognitive : float, optional
            The cognitive coefficient (default is 1.5).
        social : float, optional
            The social coefficient (default is 1.5).
        plot : bool, optional
            Whether to plot the progress of the algorithm (default is False).
    
        Returns
        -------
        best_solution : np.ndarray
            The best solution found by the algorithm. Shape: (n_dimensions,)
        best_objective : float
            The objective value of the best solution.
        solution_history : np.ndarray
            The history of solutions over generations. Shape: (n_dimensions, num_gen, pop_size)
        objective_history : np.ndarray
            The history of objective values over generations. Shape: (1, num_gen, pop_size)
        best_solution_history : np.ndarray
            The history of the best solutions over generations. Shape: (n_dimensions, num_gen)
        best_objective_history : np.ndarray
            The history of the best objective values over generations. Shape: (1, num_gen)
        """
        
        # TODO vérif cohérence pop_size et init_pop_mean / init_pop_std
    
        # Solution history is a 3D array:
        #   - 1st dimension: the solutions dimension
        #   - 2nd dimension: the generation number (or iteration number)
        #   - 3rd dimension: the individuals dimension
        # Thus with a run of 100 iterations, a 2D objective function and a population size of 30, solution_history will have a shape of (2, 100, 30)
        solution_history = np.zeros((len(init_pop_mean), num_gen, pop_size))
    
        # Objective history is a 3D array:
        #   - 1st dimension: the number of objectives to optimize (by default 1)
        #   - 2nd dimension: the generation number (or iteration number)
        #   - 3rd dimension: the individuals dimension
        # Thus with a run of 100 iterations, a mono objective function and a population size of 30, solution_history will have a shape of (1, 100, 30)
        objective_history = np.zeros((1, num_gen, pop_size))
    
        # Initialisation de la population
        population = np.random.normal(init_pop_mean, init_pop_std, (pop_size, len(init_pop_mean)))
        velocities = np.random.normal(0, 1, (pop_size, len(init_pop_mean)))
        personal_best_positions = population.copy()
        personal_best_scores = np.array([objective_function(ind) for ind in population])
        global_best_position = personal_best_positions[np.argmin(personal_best_scores)]
        global_best_score = np.min(personal_best_scores)
    
        for generation in range(num_gen):
            # Évaluation
            fitness = np.array([objective_function(ind) for ind in population])
    
            # Mise à jour des meilleures positions personnelles
            better_mask = fitness < personal_best_scores
            personal_best_positions[better_mask] = population[better_mask]
            personal_best_scores[better_mask] = fitness[better_mask]
    
            # Mise à jour de la meilleure position globale
            if np.min(fitness) < global_best_score:
                global_best_position = population[np.argmin(fitness)]
                global_best_score = np.min(fitness)
    
            # Enregistrement de la population et des valeurs d'objectifs
            solution_history[:, generation, :] = population.T
            objective_history[0, generation, :] = fitness
    
            # Mise à jour des vitesses et des positions
            r1 = np.random.rand(pop_size, len(init_pop_mean))
            r2 = np.random.rand(pop_size, len(init_pop_mean))
            velocities = (inertia * velocities +
                          cognitive * r1 * (personal_best_positions - population) +
                          social * r2 * (global_best_position - population))
            population = population + velocities
    
        # Historique du meilleur individu pour chaque génération
        best_solution_index_history = objective_history.argmin(axis=2).flatten()
        best_objective_history = objective_history[0, np.arange(num_gen), best_solution_index_history]
        best_solution_history = solution_history[:, np.arange(num_gen), best_solution_index_history]

        result = PSOResult(
            best_solution=global_best_position,
            best_objective=global_best_score,
            solution_history=solution_history,
            objective_history=objective_history,
            best_solution_history=best_solution_history,
            best_objective_history=best_objective_history
        )

        return result


"""PyEALib

An open source Python library for mathematical optimization

Note:

    This project is in beta stage.

Viewing documentation using IPython
-----------------------------------
To see which functions are available in `ealib`, type ``ealib.<TAB>`` (where
``<TAB>`` refers to the TAB key), or use ``ealib.*get_version*?<ENTER>`` (where
``<ENTER>`` refers to the ENTER key) to narrow down the list.  To view the
docstring for a function, use ``ealib.get_version?<ENTER>`` (to view the
docstring) and ``ealib.get_version??<ENTER>`` (to view the source code).
"""

import importlib.metadata

import ealib.functions
import ealib.minimizers
import ealib.utils

# PEP0440 compatible formatted version, see:
# https://www.python.org/dev/peps/pep-0440/
#
# Generic release markers:
# X.Y
# X.Y.Z # For bugfix releases  
# 
# Admissible pre-release markers:
# X.YaN # Alpha release
# X.YbN # Beta release         
# X.YrcN # Release Candidate   
# X.Y # Final release
#
# Dev branch marker is: 'X.Y.dev' or 'X.Y.devN' where N is an integer.
# 'X.Y.dev0' is the canonical version of 'X.Y.dev'

__version__ = importlib.metadata.version("ealib")

def get_version():
    return __version__

__all__ = ['TODO']